(** Sets of memory values as (mostly incompatible) DIETs and bitwords *)

type t
type valt = Z.t
type width = int

val two : valt

(** [staged_mem vs e] emits a predicate
    which determines whether [e] evals to a memval in [vs] *)
val staged_mem : t -> Target.rexpr -> Target.rexpr

val empty : width -> t
val range : width -> valt -> valt -> t
val full_range : width -> t
(** [subrange width start stop] is the set of [width]-wide words
    that start with [start] zeroes and end with [width]-[stop] ones *)
val subrange : width -> width -> width -> t
val non_zero : width -> t
val aligned_address : width -> width -> t
val singleton : width -> valt -> t

val pp : Stdlib.Format.formatter -> t -> unit

val cardinal : t -> Z.t option
val is_empty : t -> bool
val pick_one : t -> valt
val mem : valt -> t -> bool
val is_subset_of : t -> t -> bool

val union : t -> t -> t
val inter : t -> t -> t
val diff : t -> t -> t
val complement : t -> t
val arith_shift : valt -> t -> t
val slice : width -> width -> t -> t
