type 'a repr_cases =
  | TopCase
  | WordCases of (Z.t, Z.t -> 'a) Tag.KlassMap.t
  | SumCases of 'a Name.Map.t
type 'a reprs = { generic: 'a; specific: 'a repr_cases }
module type S = sig
  val pre_repr : Types.evaled_env -> Name.t -> Mem.ReprSkel.t reprs
  val repr : Types.evaled_env -> Name.t -> Mem.SolvedRepr.t reprs

  val knit : Types.evaled_env -> Name.t -> Hole.t -> Syntax.name -> Target.rexpr

  (** [frog tau h l x] émet le code qui discrimine la repr située en focus(h, x)
    suivant les cas de l; on garantit que l est exhaustive + typage se passe bien
    tau = type de x *)
  val frog : Types.evaled_env -> Name.t -> Hole.t ->
    Target.dtree Tag.Map.t -> Syntax.name -> Target.dtree
end

module type Eval = sig
  val to_pre_repr : Types.evaled_env -> Name.t -> Value.t -> SMT.problem
  val to_repr : Types.evaled_env -> Name.t -> Value.t -> Mem.t
  val match_repr : Types.evaled_env -> Name.t -> Syntax.pattern list -> Mem.t -> int
end

module type Full = sig
  include S
  include Eval
end

module Uniform : Full

module Camlike : S

module Composable : sig
  (** We need a set of discriminants of arbitrary arities
      that "cover" all tagset pairs *)
  type 'a discriminant =
    { position : Mem.Hole.t; min_width : int option;
      split : ('a, Mem.immediate option) Tag.KlassMap.t } (* None = tag placeholder *)
  module type Spec = sig
    val recompose : Types.evaled_env -> Name.t -> Hole.Head.t -> Mem.Hole.t
    val tagsets_of_word : Types.word_spec -> Z.t Tag.Klass.t list
    val discrs_of_word : Types.word_spec -> Z.t discriminant list
    val discrs_of_sum : Types.evaled_env -> Types.sum_spec -> Name.t discriminant list
  end
  module Make (R : Spec) : Full
end
