let union_fun key val1 val2 =
  assert (val1 = val2);
  Some val1

let rec valueset width : Syntax.valueset -> ValueSet.t = function
  | Full -> ValueSet.full_range width
  | Value i -> ValueSet.singleton width i
  | Range (i1, i2) -> assert false
  | Union (e1, e2) -> ValueSet.union (valueset width e1) (valueset width e2)
  | Diff (e1, e2) -> ValueSet.diff (valueset width e1) (valueset width e2)

let rec type_expr decl_env sym : Syntax.type_expr -> Types.evaled_env = function
  | Reference te ->
    let sym' = sym ^ "0" in
    let ty_env = type_expr decl_env sym' te in
    Name.Map.add sym (Types.Ref sym') ty_env
  | Tuple l ->
    let syms = List.mapi (fun i _ -> sym ^ (string_of_int i)) l in
    let ty_env = List.fold_left2
      (fun acc sym te ->
        Name.Map.union union_fun (type_expr decl_env sym te) acc)
      Name.Map.empty syms l in
    Name.Map.add sym (Types.Product syms) ty_env
  | Type_constructor { constructor; arguments = [] } ->
    (match Name.Map.find_opt constructor decl_env with
      | None -> raise (Types.Unbound_type constructor)
      | Some (Types.TyVar ty) -> Name.Map.singleton sym ty
      | Some (Types.Lam { parameters; body; body_env }) ->
        Name.Map.add sym body body_env)
  | Type_constructor { constructor; arguments } ->
    (match Name.Map.find_opt constructor decl_env with
      | None -> assert false
      | Some (TyVar ty) -> Name.Map.singleton sym ty
      | Some (Lam { parameters; body; body_env }) ->
        let env_with_args = List.fold_left2
          (fun acc param arg ->
            Name.Map.union union_fun (type_expr decl_env param arg) acc)
          body_env
          parameters arguments in
        Name.Map.add sym body env_with_args)


let declaration env Syntax.{ name; parameters; definition } : Types.evaled_env =
  match definition with
  | Syntax.Word { width; values } ->
    Name.Map.singleton name (Types.Word { width; values = valueset width values })
  | Syntax.Sum variants ->
    (* temporarily bind arg syms to dummy type vars *)
    let decl_env = List.fold_left (fun acc x -> Name.Map.add x Types.(TyVar (Var x)) acc) env parameters in
    (* also rec type sym *)
    let decl_env = Name.Map.add name Types.(TyVar (Var name)) decl_env in
    let sumspec, variant_envs = List.split @@ List.map
      (fun Syntax.{ constructor; arguments } ->
        let arg_syms = List.mapi
          (fun i _ -> name ^ constructor ^ (string_of_int i))
          arguments in
        let arg_envs = List.map2 (type_expr decl_env) arg_syms arguments in
        (constructor, arg_syms),
        List.fold_left (Name.Map.union union_fun) Name.Map.empty arg_envs)
      variants in
    let full_env = List.fold_left
      (Name.Map.union union_fun)
      Name.Map.empty variant_envs in
    Name.Map.add name (Types.Sum sumspec) full_env

