open Syntax

type t = Syntax.value =
  | VInt of Z.t
  | VRef of t
  | VTuple of t list
  | VConstr of t constructor

let rec pp fmt = function
  | VInt z -> Fmt.pf fmt "%a" Z.pp_print z
  | VRef v -> Fmt.pf fmt "&%a" pp v
  | VTuple l ->
    Fmt.pf fmt "< %a >" Fmt.(list ~sep:comma pp) l
  | VConstr { constructor; arguments } ->
    Fmt.pf fmt "%s(%a)" constructor Fmt.(list ~sep:comma pp) arguments

module Bindings = Map.Make(String)
type bindings = t Bindings.t

let rec accept_and_bind p v =
  let merge_disjoint_bindings l =
    let _, overlap = List.fold_left
      (fun (acc, res) m ->
        let keys, _ = List.split (Bindings.bindings m) in
        let overlap = List.exists (fun k -> List.mem k acc) keys in
        (keys @ acc, overlap || res))
      ([], false) l in
    if overlap then None else
    let all_binds = List.fold_left (* TODO c'est un peu nul *)
      (Bindings.union (assert false))
      Bindings.empty l in
    Some all_binds
  in
  match p, v with
  | PAny, _ -> Some (Bindings.empty)
  | PVar x, v -> Some (Bindings.singleton x v)
  | PInt c, VInt z when z = c -> Some (Bindings.empty)
  | PRef p, VRef v -> accept_and_bind p v
  | PTuple ps, VTuple vs ->
    let binds = List.map2 accept_and_bind ps vs in
    if List.exists ((=)None) binds then None else
    let binds = List.filter_map (fun b -> b) binds in
    merge_disjoint_bindings binds
  | PConstr { constructor = k; arguments = l },
    VConstr { constructor = k'; arguments = l' } when k' = k ->
    let binds = List.map2 accept_and_bind l l' in
    if List.exists ((=)None) binds then None else
    let binds = List.filter_map (fun b -> b) binds in
    merge_disjoint_bindings binds
  | POr (p1, p2), v ->
    let accept_p1 = accept_and_bind p1 v in
    (match accept_p1 with
     | None -> accept_and_bind p2 v
     | Some _ -> accept_p1)
  | _ -> None

let accept_bind_ith { clauses; _ } v =
  List.fold_left (fun (cur_match, idx) cl -> match cur_match with
    | Some _ -> (cur_match, idx + 1)
    | None -> (match accept_and_bind cl.pattern v with
      | None -> None, idx + 1
      | Some bindings -> Some (idx, cl.expr, bindings), idx + 1))
  (None, 0) clauses
